from funcy import rcompose, mapcat
from ..nodes import User, Media
from .common import accepts


@accepts(Media)
def authors(bot, nodes,  args):


    process = lambda media: media.get_author(bot)
    result = mapcat(process, nodes)

    return result, bot.last



from functools import partial

from .authors import authors
from .followers import followers
from .following import following
from .user_feed import user_feed
from .hashtag_feed import hashtag_feed
from .likers import likers
from .hashtags import hashtags
from .usertags import usertags
from .geotag import geotag
from .geotag_feed import geotag_feed
from .filter import filter
from .user_stories import user_stories

from .like import like
from .follow import follow
from .unfollow import unfollow
from .text import text
from .upload import upload
from .comment import comment
from .delete import delete
from .scrape import scrape
from .set_profile import set_profile
from .shuffle import shuffle
from .print import _print

from .evaluate import evaluate

methods = dict(
   authors=authors,
   followers=followers,
   likers=likers,
   following=following,
   user_feed=user_feed,
   hashtag_feed=hashtag_feed,
   hashtags=hashtags,
   usertags=usertags,
   geotag=geotag,
   geotag_feed=geotag_feed,
   filter=filter,
   user_stories=user_stories,

   follow=follow,
   unfollow=unfollow,
   like=like,
   text=text,
   upload=upload,
   comment=comment,
   delete=delete,
   scrape=scrape,
   set_profile=set_profile,
   shuffle=shuffle,
   print=_print,

   evaluate=evaluate,
)

#
# def make_methods(bot):
#     return {key: partial(value, bot) for (key, value) in methods.items()}



from threading import _newname, current_thread, Event, _sys, _dangling, Thread


def start(threads):
    [thread.setDaemon(True) for thread in threads]
    [thread.start() for thread in threads]
    return threads


def wait(threads):
    [thread.join() for thread in threads]
    return threads


def reset():
    return []


def custom_init(self, group=None, target=None, name=None,
                args=(), kwargs=None, *, daemon=None):

    self._data = {}

    assert group is None, "group argument must be None for now"
    if kwargs is None:
        kwargs = {}
    self._target = target
    self._name = str(name or _newname())
    self._args = args
    self._kwargs = kwargs
    if daemon is not None:
        self._daemonic = daemon
    else:
        self._daemonic = current_thread().daemon
    self._ident = None
    self._tstate_lock = None
    self._started = Event()
    self._is_stopped = False
    self._initialized = True
    self._stderr = _sys.stderr
    _dangling.add(self)


def set_data(self, value):
    self._data = value


def get_data(self):
    return self._data


Thread.__init__ = custom_init
Thread.set_data = set_data
Thread.get_data = get_data



from .node import Node


"""
{
    id
    latest_reel_media
    expiring_at
    seen
    can_reply
    can_reshare
    reel_type
    user {
        pk
        username
        full_name
        is_private
        profile_pic_url
        profile_pic_id
        friendship_status {
            following
            followed_by
            blocking
            muting
            is_private
            incoming_request
            outgoing_request
            is_bestie
        }
        is_verified
    }
    items[9] {
        taken_at
        pk
        id
        device_timestamp
        media_type
        code
        client_cache_key
        filter_type
        image_versions2 {
            candidates[2] {
                width
                height
                url
            }
        }
        original_width
        original_height
        caption_position
        is_reel_media
        is_dash_eligible
        video_dash_manifest
        video_codec
        number_of_qualities
        video_versions[3] {
            type
            width
            height
            url
            id
        }
        has_audio
        video_duration
        user {
            pk
            username
            full_name
            is_private
            profile_pic_url
            profile_pic_id
            is_verified
            has_anonymous_profile_picture
            reel_auto_archive
            is_unpublished
            is_favorite
        }
        caption {
            _
        }
        caption_is_edited
        photo_of_you
        can_viewer_save
        organic_tracking_token
        expiring_at
        imported_taken_at
        can_reshare
        story_hashtags[1] {
            x
            y
            z
            width
            height
            rotation
            is_pinned
            is_hidden
            hashtag {
                name
                id
            }
        }
        supports_reel_reactions
        show_one_tap_fb_share_tooltip
        has_shared_to_fb
    }
    prefetch_count
    has_besties_media
    media_count
    status
}
"""
attributes = lambda x: (x._id, x._data)


class Story(Node):
    __slots__ = ['_id', '_data']

    def __init__(self, *, generic=None, id=None, data=None,):
        self._id = id
        self._data = data

        if generic:
            self._data = generic

    def __repr__(self):
            url = self.get_url()
            return 'Story(id=\'{}\', url=\'{}\')'.format(self._id, url)

    def get_data(self, bot):
        return self._data

    def get_expiring_at(self):
        data = self._data
        if 'expiring_at' in data:
            return data['expiring_at']
        else:
            return None

    def get_taken_at(self):
        data = self._data
        if 'taken_at' in data:
            return data['taken_at']
        else:
            return None

    def get_url(self):
        data = self._data
        if 'video_versions' and 'video_duration' in data:
            return data['video_versions'][0]['url']
        elif 'image_versions2' in data:
            return data['image_versions2']['candidates'][0]['url']
        else:
            return None



from typing import List
from funcy import  rcompose, raiser, ignore
from operator import itemgetter
import time

from .common import accepts, today, tap
from ..nodes import Node, User, Arg




@accepts(Arg)
def set_profile(bot, nodes,  args):


    @ignore((KeyError, AttributeError), None)
    def pick(key):
        return args[key]

    mode = pick('mode')
    edits = {
        'external_url': pick('external_url'),
        'phone_number': pick('phone_number'),
        'username': pick('username'),
        'full_name': pick('full_name'),
        'biography': pick('biography'),
        'email': pick('email'),
        'gender': pick('gender'),
    }

    if mode:
        if mode == 'public':
            bot.api.set_public_account()
        elif mode == 'private':
            bot.api.set_private_account()
        else:
            bot.logger.error('{} is not either "public" or "private"'.format(mode))

    # TODO set default values in edits
    if any([value for value in edits.values()]):
        bot.api.get_profile_data()

        previous_values = bot.last['user']
        new_values = {key: value for key, value in edits.items() if value}

        edits = {**previous_values, **new_values}
        bot.api.edit_profile(**edits)

    bot.logger.info('changed profile values')

    return [], bot.last



from colorama import init, Fore

from string import Formatter

def get_field_value(field_name, mapping):
    try:
        def recursive_get(field_name, mapping):

                if '.' not in field_name:
                    return mapping[field_name], True
                else:
                    *attrs, = field_name.split('.')
                    return recursive_get(".".join(attrs[1:]), mapping[attrs[0]])
        return recursive_get(field_name, mapping)

    except:
        # traceback.print_exc()
        return field_name, False




def str_format_map(format_string, mapping):
    f = Formatter()
    parsed = f.parse(format_string)
    output = []
    for literal_text, field_name, format_spec, conversion in parsed:
        conversion = '!' + conversion if conversion is not None else ''
        format_spec = ':' + format_spec if format_spec else ''
        if field_name is not None:
            field_value, found = get_field_value(field_name, mapping)
            if not found:
                text = '{{{}{}{}}}'.format(field_value,
                                           conversion,
                                           format_spec)
            else:
                format_string = '{{{}{}}}'.format(conversion, format_spec)
                text = format_string.format(field_value)
        output.append(literal_text + text)
        text = ''
    return ''.join(output)

def populate_object(script, data={}):
    def recursive_populate(script, data):
        variables = data #safedotdict(**data)
        if isinstance(script, dict):
            for (key, value) in script.items():
                if isinstance(value, str):
                    if '{' in value and '}' in value:
                        try:
                            new_value = str_format_map(value, variables)
                            script[key] = new_value
                        except Exception as e:
                            init(autoreset=True)
                            print(Fore.RED + 'error in {},\n{}'.format(value, e))
                else:
                    recursive_populate(script[key], data)
        else:
            return

    script_copy = dict(**script)
    recursive_populate(script_copy, data)
    return script_copy

def populate_string( yaml_string, data={}):
    """
    max one {{  }} per line!
    """

    def replace_in_line(line):
        if '{{' in line and '}}' in line:
            begin = line.index('{{')
            end = line.index('}}', begin)
            variable_name = line[begin:end].strip().replace('{{','').replace('}}','').strip()
            if variable_name in data:
                return (
                    line[:begin].replace('{{','').replace('}}','') +
                    str(data[variable_name]) +
                    line[end:].replace('}}','').replace('{{','')
                )
            else:
                return line
        else:
            return line

    new_lines = map(replace_in_line, yaml_string.splitlines())
    return '\n'.join(new_lines)



"""
HTML logger inspired by the Horde3D logger.
Usage:
 - call setup and specify the filename, title, version and level
 - call dbg, info, warn or err to log messages.
"""
import logging
import time

#: HTML header (starts the document
_START_OF_DOC_FMT = """
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<title>%(title)s</title>
<style type="text/css">
body, html {
background: #000000;
width: 1000px;
font-family: Arial;
font-size: 16px;
color: #C0C0C0;
}
h1 {
color : #FFFFFF;
border-bottom : 1px dotted #888888;
}
pre {
font-family : arial;
margin : 0;
}
.box {
border : 1px dotted #818286;
padding : 5px;
margin: 5px;
width: 950px;
background-color : #292929;
}
.err {
color: #EE1100;
font-weight: bold
}
.warn {
color: #FFCC00;
font-weight: bold
}
.info {
color: #C0C0C0;
}
.debug {
color: #CCA0A0;
}
</style>
</head>
<body>
<h1>%(title)s</h1>
<h3>%(version)s</h3>
<div class="box">
<table>
"""

_END_OF_DOC_FMT = """</table>
</div>
</body>
</html>
"""

_MSG_FMT = """
<tr>
<td width="100">%(time)s</td>
<td class="%(class)s"><pre>%(msg)s</pre></td>
<tr>
"""


class HTMLFileHandler(logging.FileHandler):
    """
    File handler specialised to write the start of doc as html and to close it
    properly.
    """

    def __init__(self, title,  file, *args, version='1.0.0', mode='w', ):
        super().__init__(file, *args, mode=mode)
        assert self.stream is not None
        # Write header
        self.stream.write(_START_OF_DOC_FMT % {"title": title,
                                               "version": version})

    def close(self):
        # finish document
        self.stream.write(_END_OF_DOC_FMT)
        super().close()


class HTMLFormatter(logging.Formatter):
    """
    Formats each record in html
    """
    CSS_CLASSES = {'WARNING': 'warn',
                   'INFO': 'info',
                   'DEBUG': 'debug',
                   'CRITICAL': 'err',
                   'ERROR': 'err'}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start_time = time.time()

    def format(self, record):
        try:
            class_name = self.CSS_CLASSES[record.levelname]
        except KeyError:
            class_name = "info"

        t = time.time() - self._start_time

        # handle '<' and '>' (typically when logging %r)
        msg = record.msg
        if not isinstance(msg, str):
            msg = str(msg)
        msg = msg.replace("<", "&#60")
        msg = msg.replace(">", "&#62")

        return _MSG_FMT % {"class": class_name, "time": "%.4f" % t,
                           "msg": msg}


# class HTMLLogger(logging.Logger):
#     """
#     Log records to html using a custom HTML formatter and a specialised
#     file stream handler.
#     """
#
#     def __init__(self,
#                  name="html_logger",
#                  level=logging.DEBUG,
#                  filename="log.html", mode='w',
#                  title="HTML Logger", version="1.0.0"):
#         super().__init__(name, level)
#         f = HTMLFormatter()
#         h = HTMLFileHandler(title, version, filename, mode)
#         h.setFormatter(f)
#         self.addHandler(h)
#
#
# #: Global logger instance
# _logger = None
#
#
# def setup(title, version, filename="log.html", mode='w', level=logging.DEBUG):
#     """
#     Setup the logger
#     :param title: Title of the html document
#     :param version: Framework/lib/app version
#     :param filename: output filename. Default is "log.html"
#     :param mode: File open mode. Default is 'w'
#     :param level: handler output level. Default is DEBUG
#     """
#     global _logger
#     if _logger is None:
#         _logger = HTMLLogger(filename=filename, mode=mode, title=title,
#                              version=version, level=level)
#
#
# def dbg(msg):
#     """
#     Logs a debug message
#     """
#     global _logger
#     _logger.debug(msg)
#
#
# def info(msg):
#     """
#     Logs an info message
#     """
#     global _logger
#     _logger.info(msg)
#
#
# def warn(msg):
#     """
#     Logs a warning message
#     """
#     global _logger
#     _logger.warning(msg)
#
#
# def err(msg):
#     """
#     Logs an error message
#     """
#     global _logger
#     _logger.error(msg)
#
#
# # Example of usage
# if __name__ == "__main__":
#     setup("Example", "1.0")
#     dbg("A debug message")
#     info("An information message")
#     warn("A warning message")
#     time.sleep(1)
#     err("An error message")




from .node import Node
from .user import User
from .hashtag import Hashtag
from .geotag import Geotag
from .arg import Arg
from .media import Media
from .story import Story


node_classes = dict(
    node=Node,
    user=User,
    media=Media,
    story=Story,
    hashtag=Hashtag,
    geotag=Geotag,
    arg=Arg,
)



from psutil import net_io_counters
from threading import Thread
import time
from .safe_print import safe_print

def bytes2human(n):
    """
    >>> bytes2human(10000)
    '9.8 K'
    >>> bytes2human(100001221)
    '95.4 M'
    """
    symbols = ('KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if n >= prefix[s]:
            value = float(n) / prefix[s]
            return '%.2f %s' % (value, s)
    return '%.2f B' % (n)


class Network_logger(Thread):
    def __init__(self):
        super().__init__()
        self.elapsed_minutes = 0
        self.recv = 0
        self.sent = 0

    def run(self):
        try:
            while True:
                net_io_counters.cache_clear()
                net = net_io_counters(nowrap=True)
                self.recv = net.bytes_recv
                self.sent =  net.bytes_recv
                safe_print('[after {} min received {} and sent {}]'.format(self.elapsed_minutes, bytes2human(self.recv), bytes2human(self.sent)))
                time.sleep(60)
                self.elapsed_minutes += 1
        except KeyboardInterrupt:
            raise



from funcy import rcompose, take, filter as _filter
from itertools import islice
from ..make_predicate import make_predicate
from ..nodes import Node
from .common import accepts


@accepts(Node)
def filter(bot, nodes,  args) -> Node:

    if args:
        predicate = make_predicate(args, bot)
        nodes = _filter(predicate, nodes)
    else:
        nodes = bot.filter(nodes)


    return nodes, bot.last



from funcy import rcompose
from ..nodes import Geotag, Media
from .common import accepts


@accepts(Media)
def geotag(bot, nodes,  args) -> Geotag:

    process = lambda media: media.get_geotag(bot)
    result = map(process, nodes)

    return result, bot.last




from typing import List
from funcy import  rcompose, mapcat
from ..nodes import User, Media
from .common import accepts, cycled_api_call, tap

@accepts(User)
def user_feed(bot, nodes,  args) -> List[Media]:
    amount = args.get('amount')

    pack_media = lambda data: Media(id=data['pk'], data=data)

    process = rcompose(
        lambda user: user.id if user.id else user.get_id(bot),
        lambda id: cycled_api_call(amount, bot, bot.api.get_user_feed, id, 'items'),
        lambda items: map(pack_media, items),
        # lambda x: tap(x, lambda: print(x)),

    )

    result = mapcat(process, nodes)

    return result, bot.last



from .node import Node

"""
    pk
    name
    address
    city
    short_name
    lng
    lat
    external_source
    facebook_places_id
"""

"""
    "pk": 186886085560507,
    "name": "Kartepe",
    "address": "",
    "city": "",
    "short_name": "Kartepe",
    "lng": 30.195368,
    "lat": 40.675953,
    "external_source": "facebook_places",
    "facebook_places_id": 186886085560507
"""

class Geotag(Node):
    __slots__ = ['_name', '_id', '_data']
    def __init__(self, *, generic=None, name=None, id=None, data=None):

        self._name = name
        self._id = id
        self._data = data

        if generic:
            self._name = generic

    def __repr__(self):
        name, id, data = attributes(self)

        if name:
            return 'Geotag(name=\'{}\')'.format(name)
        elif id:
            return 'Geotag(id=\'{}\')'.format(id)
        else:
            return 'Geotag(...)'

    @property
    def name(self):
        name, id, data = attributes(self)

        if name:
            return name
        elif data:
            return data['name']
        else:
            return False

    @property
    def id(self):
        name, id, data = attributes(self)
        if id:
            return id
        elif data:
            return data['pk']
        else:
            return False

    def get_data(self, bot):
        name, id, data = attributes(self)
        if data:
            if 'pk' in data:
                self._id = data['pk']
                return self.get_data(bot)
            else:
                return False
        elif name:
            bot.api.search_location(query=name)
            try:
                bot.api.search_location(query=name)
                data = bot.last['items'][0]['location']
                self._data = data
                return data
            except KeyError:
                return False
        elif 'lat' in data and 'lng' in data:
            bot.api.search_location(lat=data['lat'], lng=data['lng'])
            try:
                bot.api.search_location(query=name)
                data = bot.last['items'][0]['location']
                self._data = data
                return data
            except KeyError:
                return False
        else:
            return False


    def get_id(self, bot):
        name, id, data = attributes(self)
        if id:
            return id
        elif data:
            if 'pk' in data:
                return data['pk']
        elif name:
            data = self.get_data(bot)
            return data['pk']
        else:
            return False

    def get_facebook_id(self, bot):
        _, _, data = attributes(self)
        if 'facebook_id' in data:
            return data['facebook_id']
        else:
            data = self.get_data(bot)
            return data['facebook_id']

    def get_coordinates(self, bot):
        _, _, data = attributes(self)
        if 'lat' in data and 'lng' in data:
            return data['lat'], data['lng']
        else:
            data = self.get_data(bot)
            return data['lat'], data['lng']

attributes = lambda x: (x._name, x._id, x._data)



from .methods import methods
from .nodes import node_classes, Media, User, Arg




def nodes_edges(body):

    nodes = []
    edges = []

    if 'nodes' in body:
        nodes += body['nodes']

    if 'edges' in body:
        for edge in body['edges']:

            if isinstance(edge, dict):
                type = list(edge.keys())[0]
                args = list(edge.values())[0]
                edges += [dict(type=type, args=args)]
            else:
                edges += [dict(type=edge, args={})]
    else:
        raise Exception('in every action there must be edges')

    edges += [dict(type='evaluate', args={})]

    if 'name' in body:
        name = body['name']

    else:
        name='not named'

    for n, _ in enumerate(edges):
        edges[n]['name'] = name

    nodes = initialize_nodes(nodes, edges, body)
    return nodes, edges






def initialize_nodes(nodes, edges, data):
    if 'from_type' in data:
        Class = node_classes[data['from_type']]
    else:
        first_method = methods.get(edges[0]['type'], None)
        if not first_method:
            raise Exception('can\'t find {} edge in available edges methods')
        Class = first_method.accepts
        Class = Class if Class.__name__ != 'Node' else \
            Media if 'instagram.com' in nodes[0] else Arg

    return [Class(generic=value) for value in nodes]




from typing import List
from funcy import  rcompose, mapcat
from itertools import tee
import json

from ..nodes import User, Media, Story
from .common import accepts, cycled_api_call, tap

@accepts(User)
def user_stories(bot, nodes,  args) -> List[Story]:

    amount = args.get('amount')
    pack_story = lambda data: Story(id=data['pk'], data=data)
    # unmasked = lambda: unmask(bot.last)
    # log_unmasked = lambda: bot.logger.warn(unmasked())
    # bot.logger.warn([x for x in nodes])


    process = rcompose(
        lambda user: user.id if user.id else user.get_id(bot),
        # lambda id: tap(id, lambda: bot.api.get_user_stories(id)),
        # lambda x: tap(x, lambda: print(x)),
        lambda id: get_stories(bot, id),
        # lambda x: tap(x, lambda: print(x)),
        lambda gen: map(pack_story, gen)
    )

    stories = mapcat(process, nodes)

    return stories, bot.last

def get_stories(bot, user_id):
    bot.api.get_user_stories(user_id)
    if 'reel' in bot.last:
        yield from bot.last['reel']['items']
    else:
        yield from []



from funcy import rcompose, mapcat
from itertools import islice
from ..nodes import Hashtag, Media
from .common import accepts


@accepts(Media)
def hashtags(bot, nodes,  args):

    amount = args.get('amount')
    process = lambda media: media.get_hashtags(bot)
    result = mapcat(process, nodes)
    result = islice(result, amount)

    return result, bot.last




from ..nodes import Media, User, Hashtag, Geotag, Node

def not_in_cache(bot, node,table=None, specifier=None,):

    with bot.cache as cache:

        if table and not specifier:
            if cache[table].find_one(identifier=node.id, ):
                return False
        elif table and specifier:
            if cache[table].find_one(identifier=node.id,  specifier=specifier):
                return False

        if isinstance(node, Media):
            kwargs = dict(identifier=node.id, type='media', )
            if specifier: kwargs['specifier'] = specifier

            if cache['liked'].find_one(**kwargs):
                return False
            elif cache['commented'].find_one(**kwargs):
                return False
            elif cache['reported'].find_one(**kwargs):
                return False
            else:
                return True


        elif isinstance(node, User):
            kwargs = dict(identifier=node.id, type='user', )
            if specifier: kwargs['specifier'] = specifier

            if cache['followed'].find_one(**kwargs):
                return False
            elif cache['blocked'].find_one(**kwargs):
                return False
            elif cache['texted'].find_one(**kwargs):
                return False
            else:
                return True

        elif isinstance(node, Hashtag):
            return True

        elif isinstance(node, Geotag):
            return True

        elif isinstance(node, Node):
            return True

        else:
            return False






TOTAL = {'likes': 0,
         'unlikes': 0,
         'follows': 0,
         'unfollows': 0,
         'comments': 0,
         'blocks': 0,
         'unblocks': 0,
         'texts': 0,
         'archived': 0,
         'unarchived': 0,
         'uploads': 0
         }


DELAY = {'like': 10,
         'unlike': 10,
         'follow': 5,
         'unfollow': 3,
         'comment': 60,
         'block': 30,
         'unblock': 30,
         'text': 5,
         'upload': 30,
         'delete': 30,
         'error': 30,

         'usual': 4,
         'getter': 70
    }


MAX_PER_DAY = {'likes': 1000,
               'unlikes': 1000,
               'follows': 350,
               'unfollows': 350,
               'comments': 100,
               'blocks': 100,
               'unblocks': 100,
               'texts': 300
               }





# from .make_task import make_task, partitionate
from .nodes_edges import nodes_edges
from .make_bots import make_bots
from .populate import populate_object, populate_string
from .reducer import Reducer, reducer
from functools import reduce
# from .threads import start, wait
import traceback

from ruamel.yaml import YAML

yaml = YAML()

def execute(script, variables={}) -> [dict]:

    script = obj_from_yaml(script, variables)

    bot = make_bots(script)[0]

    script_name = script['name'] if 'name' in script else 'unnmaed script'
    bot.logger.info(f'# SCRIPT {script_name}')

    data = []

    try:
        for action in script['actions']:
            action_name = action['name'] if 'name' in action else 'unnmaed action'
            bot.logger.info(f'# ACTION {action_name}')

            nodes, edges = nodes_edges(action)
            state = dict(nodes=nodes, bot=bot, data=dict(), errors=[])
            end_state = reduce(reducer, edges, state)
            data += end_state['data']

    # try:
    #     for action in script['actions']:
    #
    #
    #
    #         threads = []
    #         task = make_task(action)
    #         name = task['name']
    #
    #         for (task, bot) in partitionate(task, bots):
    #             state = dict(nodes=task.nodes, bot=bot, data=dict(), errors=[])
    #             threads += [Reducer(state, task.edges)]
    #             # bot.logger.debug('edges : {}, nodes: {}'.format(edges, list(state['nodes'])))
    #
    #
    #         threads = start(threads)
    #         threads = wait(threads)
    #
    #         data['__' + interaction + '_interaction__'] = [thread.get_data() for thread in threads]
    #         # {'thread' + thread.name: thread.get_data() for thread in threads}


    except KeyboardInterrupt:
        bot.logger.warn('keyboard interrupt')
        raise

    except Exception as exc:
        print(
            exc.__class__.__name__,
            ':',
            exc,
            '\n',
            '\n'.join(traceback.format_exc().split('\n'))
        )
        raise

    finally:
        return data

def locate_variable(script):
    begin = script.index('{{')
    end = script.index('}}', begin )
    return script[begin:end].replace('{{', '').strip()


def obj_from_yaml(script, variables):
    if isinstance(script, str):
        script = populate_string(script, variables)
        if '{{' in script:
            var = locate_variable(script)
            raise Exception('yaml file needs all data to be populated: {{{{ {} }}}}'.format(var))
        return yaml.load(script)
    else:
        return populate_object(script, variables)



import traceback
from ..nodes import Node
from .common import accepts


@accepts(Node)
def evaluate(bot, nodes,  args) -> Node:

    while True:
        try:
            
            try:
                next(nodes)

            except TypeError:
                next(iter(nodes))

        except StopIteration:
            break
        except:
            bot.logger.error('there was an error:\n{}'.format(traceback.format_exc()))
            bot.sleep('error')
            pass

    # bot.logger.warn(nodes[:3])

    return nodes, bot.last



import sys
import json
from colorama import init, Fore
from .execute import execute
# from .debug import unmask



"""
python -m src /etc/template.yml --data-file /etc/data.json

"""
import click
# from instamob import execute

if __name__ == '__main__':
    @click.command()
    @click.argument('file', type=str)
    @click.option('--data-file', help='data variables to add in the template file')
    def main(ship, template_path, data_path):
        data= json.load(data_path,)
        template = load(template_path)

        out = execute(template, data)
        init()
        print(Fore.YELLOW + out)


def load(file):
    with open(file, 'r') as f:
        return f.read()



"""
Configuration file of the project
"""
# Config variables taken from
# https://github.com/ping/instagram_private_api/blob/master/instagram_private_api/constants.py
API_URL = 'https://i.instagram.com/api/v1/'
USER_AGENT_BASE = (
    'Instagram {instagram_version} '
    'Android ({android_version}/{android_release}; '
    '{dpi}; {resolution}; {manufacturer}; '
    '{device}; {model}; {cpu}; en_US)')
SIG_KEY_VERSION = '4'
IG_SIG_KEY = '99e16edcca71d7c1f3fd74d447f6281bd5253a623000a55ed0b60014467a53b1'

# Request variables taken from
# https://github.com/ping/instagram_private_api/blob/422d61f0a8cc9de3d5a0e78bcba53751c44e5d63/instagram_private_api/client.py#L375
REQUEST_HEADERS = {
    'Connection': 'Keep-Alive',
    'X-IG-Capabilities': '3brTBw==',  # base64.b64encode(struct.pack('<i', 131316445)).decode('ascii')
    'X-IG-Connection-Type': 'WIFI',
    'X-FB-HTTP-Engine': 'Liger',
    'Accept': '*/*',
    'Accept-Encoding': 'gzip,deflate',
    'Accept-Language': 'en-US',
    'Content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
    'Cookie2': '$Version=1',
}

# Moved to the end as it's quite long
EXPERIMENTS = (
    'ig_android_disk_cache_match_journal_size_to_cache_max_count,'
    'ig_android_ad_move_carousel_indicator_to_ufi_universe,'
    'ig_android_universe_video_production,'
    'ig_android_live_follow_from_comments_universe,'
    'ig_android_ad_watchandinstall_universe,ig_android_live_analytics,'
    'ig_android_video_captions_universe,ig_android_offline_location_feed,'
    'ig_android_ontact_invite_universe,'
    'ig_android_insta_video_reconnect_viewers,'
    'ig_android_live_broadcast_blacklist,'
    'ig_android_checkbox_instead_of_button_as_follow_affordance_universe,'
    'ig_android_ufi_redesign_video_social_context,'
    'ig_android_stories_surface_universe,'
    'ig_android_verified_comments_universe,'
    'ig_android_preload_media_ahead_in_current_reel,'
    'android_instagram_prefetch_suggestions_universe,'
    'ig_android_direct_inbox_tray_suggested_user_universe,'
    'ig_android_direct_blue_tab,ig_android_light_status_bar_universe,'
    'ig_android_asset_button_new_content_animation,'
    'ig_android_async_network_tweak_universe,'
    'ig_android_react_native_lazy_modules_killswitch,'
    'ig_android_instavideo_remove_nux_comments,ig_video_copyright_whitelist,'
    'ig_android_ad_sponsor_label_story_top_design_universe,'
    'ig_android_business_action,ig_android_direct_link_style,'
    'ig_android_live_heart_enhancements_universe,'
    'ig_android_preload_item_count_in_reel_viewer_buffer,'
    'ig_android_auto_retry_post_mode,ig_android_fix_render_thread_crash,'
    'ig_android_shopping,ig_fbns_preload_default,'
    'ig_android_gesture_dismiss_reel_viewer,'
    'ig_android_ad_logger_funnel_logging_universe,'
    'ig_android_direct_links,ig_android_links_receivers,'
    'ig_android_ad_impression_backtest,'
    'ig_android_offline_freshness_toast_10_12,'
    'ig_android_invites_without_token_universe,'
    'ig_android_immersive_viewer,ig_android_mqtt_skywalker,'
    'ig_fbns_push,ig_android_react_native_universe,'
    'ig_android_special_brush,ig_android_live_consumption_abr,'
    'ig_android_story_viewer_social_context,'
    'ig_android_explore_verified_badges_stories_universe,'
    'ig_android_video_loopcount_int,'
    'ig_android_enable_main_feed_reel_tray_preloading,'
    'ig_android_ad_watchbrowse_universe,ig_android_react_native_ota,'
    'ig_android_discover_people_icon_in_others_profile,'
    'ig_android_log_mediacodec_info,'
    'ig_android_enable_back_navigation_nux_universe,'
    'ig_android_cold_start_feed_request,ig_video_use_sve_universe,'
    'ig_android_offline_explore_10_14,'
    'ig_android_stories_teach_gallery_location,'
    'ig_android_http_stack_experiment_2017,'
    'ig_android_stories_device_tilt,'
    'ig_android_pending_request_search_bar,'
    'ig_android_fb_topsearch_sgp_fork_request,'
    'ig_android_animation_perf_reporter_timeout,ig_android_new_block_flow,'
    'ig_android_direct_address_links,'
    'ig_android_share_profile_photo_to_feed_universe,'
    'ig_android_stories_private_likes,ig_android_text_background,'
    'ig_android_stories_video_prefetch_kb,ig_android_su_activity_feed,'
    'ig_android_live_stop_broadcast_on_404,ig_android_render_iframe_interval,'
    'ig_android_boomerang_entry,ig_android_camera_shortcut_universe,'
    'ig_android_fetch_fresh_viewer_list,'
    'ig_android_ad_media_url_logging_universe,'
    'ig_android_phone_confirm_rate_limit_language_universe,'
    'ig_android_keep_http_cache_on_user_switch,'
    'ig_android_facebook_twitter_profile_photos,'
    'ig_android_full_user_detail_endpoint,'
    'ig_android_direct_sqlite_universe,'
    'ig_android_family_bridge_share,'
    'ig_android_search,'
    'ig_android_insta_video_consumption_titles,'
    'ig_android_live_notification_control,'
    'ig_android_camera_universe,ig_android_instavideo_audio_only_mode,'
    'ig_android_live_video_reactions_consumption_universe,'
    'ig_android_swipe_fragment_container,'
    'ig_creation_growth_holdout,ig_android_live_save_to_camera_roll_universe,'
    'ig_android_ad_cta_redesign_universe,ig_android_sticker_region_tracking,'
    'ig_android_unified_inbox,ig_android_offline_main_feed_10_11,'
    'ig_android_chaining_teaser_animation,'
    'ig_android_business_conversion_value_prop_v2,'
    'ig_android_redirect_to_low_latency_universe,'
    'ig_android_feed_header_profile_ring_universe,'
    'ig_family_bridges_holdout_universe,'
    'ig_android_following_follower_social_context,'
    'ig_android_video_keep_screen_on,ig_android_profile_photo_as_media,'
    'ig_android_insta_video_consumption_infra,'
    'ig_android_sms_consent_in_edit_profile,'
    'ig_android_infinite_scrolling_launch,'
    'ig_in_feed_commenting,ig_android_live_broadcast_enable_year_class_2011,'
    'ig_android_direct_phone_number_links,ig_android_direct_share_sheet_ring,'
    'ig_android_stories_weblink_creation,ig_android_histogram_reporter,'
    'ig_android_network_cancellation,ig_android_react_native_insights,'
    'ig_android_insta_video_audio_encoder,ig_android_family_bridge_bookmarks,'
    'ig_android_dash_for_vod_universe,'
    'ig_android_direct_mutually_exclusive_experiment_universe,'
    'ig_android_stories_selfie_sticker,'
    'ig_android_ad_add_per_event_counter_to_logging_event,ig_android_rtl,'
    'ig_android_direct_send_auto_retry,'
    'ig_android_direct_video_autoplay_scroll,'
    'ig_android_promote_from_profile_button,ig_android_share_spinner,'
    'ig_android_profile_share_username,'
    'ig_android_sidecar_edit_screen_universe,'
    'ig_promotions_unit_in_insights_landing_page,'
    'ig_android_save_longpress_tooltip,'
    'ig_android_constrain_image_size_universe,'
    'ig_android_business_new_graphql_endpoint_universe,'
    'ig_ranking_following,ig_android_universe_reel_video_production,'
    'ig_android_sfplt,ig_android_offline_hashtag_feed,'
    'ig_android_live_skin_smooth,ig_android_stories_posting_offline_ui,'
    'ig_android_direct_add_local_thread_in_inbox,'
    'ig_android_swipe_navigation_x_angle_universe,'
    'ig_android_offline_mode_holdout,ig_android_non_square_first,'
    'ig_android_insta_video_drawing,ig_android_react_native_usertag,'
    'ig_android_swipeablefilters_universe,'
    'ig_android_analytics_logger_running_background_universe,'
    'ig_android_save_all,ig_android_reel_viewer_data_buffer_size,'
    'ig_android_disk_cache_has_sanity_check,'
    'ig_direct_quality_holdout_universe,ig_android_family_bridge_discover,'
    'ig_android_react_native_restart_after_error_universe,'
    'ig_story_tray_peek_content_universe,ig_android_profile,'
    'ig_android_high_res_upload_2,ig_android_http_service_same_thread,'
    'ig_android_remove_followers_universe,ig_android_skip_video_render,'
    'ig_android_live_viewer_comment_prompt_universe,'
    'ig_android_search_client_matching,ig_explore_netego,'
    'ig_android_boomerang_feed_attribution,'
    'ig_android_explore_story_sfslt_universe,'
    'ig_android_rendering_controls,ig_android_os_version_blocking,'
    'ig_android_encoder_width_safe_multiple_16,'
    'ig_android_direct_video_autoplay,ig_android_snippets_profile_nux,'
    'ig_android_e2e_optimization_universe,ig_android_disk_usage,'
    'ig_android_save_collections,'
    'ig_android_live_see_fewer_videos_like_this_universe,'
    'ig_android_live_view_profile_from_comments_universe,'
    'ig_formats_and_feedbacks_holdout_universe,'
    'ig_fbns_blocked,ig_android_instavideo_periodic_notif,'
    'ig_android_empty_feed_redesign,ig_android_marauder_update_frequency,'
    'ig_android_suggest_password_reset_on_oneclick_login,'
    'ig_android_live_special_codec_size_list,'
    'ig_android_enable_share_to_messenger,'
    'ig_android_live_video_reactions_creation_universe,'
    'ig_android_live_hide_viewer_nux,ig_android_channels_home,'
    'ig_android_sidecar_gallery_universe,ig_android_live_using_webrtc,'
    'ig_android_insta_video_broadcaster_infra_perf,'
    'ig_android_business_conversion_social_context,'
    'android_ig_fbns_kill_switch,ig_android_retry_story_seen_state,'
    'ig_android_react_native_universe_kill_switch,'
    'ig_android_stories_book_universe,'
    'ig_android_all_videoplayback_persisting_sound,'
    'ig_android_cache_layer_bytes_threshold,'
    'ig_android_comment_deep_linking_v1,ig_android_business_promotion,'
    'ig_android_anrwatchdog,ig_android_qp_kill_switch,'
    'ig_android_ad_always_send_ad_attribution_id_universe,'
    'ig_android_2fac,ig_direct_bypass_group_size_limit_universe,'
    'ig_android_promote_simplified_flow,ig_android_share_to_whatsapp,'
    'ig_fbns_dump_ids,ig_android_ad_show_mai_cta_loading_state_universe,'
    'ig_android_skywalker_live_event_start_end,'
    'ig_android_toplive_verified_badges_universe,'
    'ig_android_live_join_comment_ui_change,'
    'ig_android_draw_button_new_tool_animation,'
    'ig_video_max_duration_qe_preuniverse,ig_android_http_stack_kz_debug,'
    'ig_request_cache_layer,ig_android_carousel_feed_indicators_universe,'
    'ig_android_new_optic,ig_android_mark_reel_seen_on_Swipe_forward,'
    'ig_fbns_shared,ig_android_capture_slowmo_mode,'
    'ig_android_save_multi_select,ig_android_mead,'
    'ig_android_video_single_surface,ig_android_offline_reel_feed,'
    'ig_android_video_download_logging,ig_android_last_edits,'
    'ig_android_exoplayer_4142,ig_android_snippets_haptic_feedback,'
    'ig_android_gl_drawing_marks_after_undo_backing,'
    'ig_android_mark_seen_state_on_viewed_impression,'
    'ig_android_live_backgrounded_reminder_universe,'
    'ig_android_disable_comment_public_test,ig_android_user_detail_endpoint,'
    'ig_android_comment_tweaks_universe,ig_android_add_to_last_post,'
    'ig_save_insights,ig_android_live_enhanced_end_screen_universe,'
    'ig_android_ad_add_counter_to_logging_event,ig_android_sidecar,'
    'ig_android_direct_split_new_message_button,ig_android_grid_video_icon,'
    'ig_android_ad_watchandlead_universe,ig_android_progressive_jpeg,'
    'ig_android_offline_story_stickers,ig_android_direct_inbox_unseen_hint,'
    'ig_android_top_live_titles_universe,'
    'ig_android_video_prefetch_for_connectivity_type,'
    'ig_android_ad_holdout_16m5_universe,'
    'ig_android_sync_on_background_enhanced,'
    'ig_android_upload_reliability_use_fbupload_lib,'
    'ig_android_samsung_app_badging,ig_android_offline_commenting,'
    'ig_android_insta_video_abr_resize,ig_android_insta_video_sound_always_on,'
    'ig_android_disable_comment'
)

SUPPORTED_CAPABILITIES = [{"name": "SUPPORTED_SDK_VERSIONS",
                           "value": "13.0,14.0,15.0,16.0,17.0,18.0,19.0,20.0,21.0,22.0,23.0,24.0,25.0,26.0,27.0,28.0,29.0,30.0,31.0,32.0,33.0,34.0,35.0,36.0,37.0,38.0,39.0,40.0,41.0,42.0,43.0,44.0,45.0,46.0,47.0,48.0,49.0,50.0,51.0,52.0,53.0"},
                          {"name": "FACE_TRACKER_VERSION", "value": "12"},
                          {"name": "segmentation", "value": "segmentation_enabled"},
                          {"name": "WORLD_TRACKER", "value": "WORLD_TRACKER_ENABLED"}]



from typing import List
from funcy import  rcompose, mapcat
from ..nodes import  Media, Hashtag
from .common import accepts, cycled_api_call

@accepts(Hashtag)
def hashtag_feed(bot, nodes,  args) -> List[Media]:
    amount = args.get('amount')

    pack_media = lambda data: Media(id=data['pk'], data=data)

    process = rcompose(
        lambda tag: tag.name,
        # lambda x: tap(x, lambda: print(bot.last)),
        lambda id: cycled_api_call(amount, bot, bot.api.get_hashtag_feed, id, 'items'),
        lambda items: map(pack_media, items),
    )

    result = mapcat(process, nodes)

    return result, bot.last




from funcy import rcompose, ignore
from .node import Node
from .user import User
from .geotag import Geotag
from .hashtag import Hashtag
import time


attributes = lambda media: (media._url, media._id, media._data)

class Media(Node):

    __slots__ = ['_url', '_id', '_data']

    def __init__(self, *, generic=False, url=False, id=False, data={}):

        self._url = url
        self._id = id
        self._data = data

        if generic:
            self._url = generic

    def __repr__(self):
        url, id, data = attributes(self)
        if url:
            return 'Media(url=\'{}\')'.format(url)
        elif id:
            return 'Media(id={})'.format(id)
        elif data:
            return 'Media(data=\'{...}\')'

    @property
    def id(self):
        url, id, data = attributes(self)
        if id:
            return id
        elif url:
            return id_from_url(url)
        else:
            raise Exception('can\'t retrieve id, not enought data')

    @property
    def url(self):
        url, id, data = attributes(self)
        if url:
            return url
        elif id:
            return url_from_id(id)
        elif data:
            return data['media_id']
        else:
            raise Exception

    def get_data(self, bot):
        url, id, data = attributes(self)
        if id:
            bot.api.media_info(id)
            bot.sleep('usual')
            if 'items' in bot.last:
                self._data = bot.last["items"][0]
                return self._data
        elif data:
            if 'media_id' in data:
                self._id = data['media_id']
                return self.get_data(bot)
            else:
                return {}
        elif url:
            id = id_from_url(url)
            self._id = id
            return self.get_data(bot)

        else:
            return {}


    def get_author(self, bot):
        _, _, data = attributes(self)
        if 'user' in data:
            user = data["user"]
            return User(
                data=user,
                id=user['pk'],
                username=user['username']
            )
        else:
            data = self.get_data(bot)
            user = data["user"]
            return User(
                data=user,
                id=user['pk'],
                username=user['username']
            )

    def get_like_count(self, bot):
        _, _, data = attributes(self)
        if 'like_count' in data:
            return data['like_count']
        else:
            data = self.get_data(bot)
            return data['like_count']

    def get_comment_count(self, bot):
        _, _, data = attributes(self)
        if 'comment_count' in data:
            return data['comment_count']
        else:
            data = self.get_data(bot)
            if 'comment_count' in data:
                return data['comment_count']
            else:
                return False

    @ignore(TypeError)
    def get_caption(self, bot):
        _, _, data = attributes(self)
        if 'caption' in data:
            return data['caption']['text']
        else:
            data = self.get_data(bot)
            return data['caption']['text']


    def get_geotag(self, bot):
        _, _, data = attributes(self)
        if 'location' in data:
            return data['location']
        else:
            data = self.get_data(bot)
            if 'location' in data:
                return  Geotag(
                    name=data['name'],
                    id=data['pk'],
                    data=data
                )
            else:
                return False

    def get_usertags(self, bot):
        _, _, data = attributes(self)
        pack_user = rcompose(
            lambda data: data['user'],
            lambda data: User(username=data['username'], id=data['pk'], data=data)
        )

        try:
            if 'usertags' in data:
                items =  data["usertags"]["in"]
                yield from (pack_user(item) for item in items)
            else:
                data = self.get_data(bot)
                if 'usertags' in data:
                    items =  data["usertags"]["in"]
                    yield from (pack_user(item) for item in items)
                else:
                    return False
        except KeyError:
            return False

    def get_hashtags(self, bot):
        text = self.get_caption(bot)
        raw_tags = set(part[1:] for part in text.split() if part.startswith('#'))
        tags = (Hashtag(name=tag) for tag in raw_tags)
        yield from tags



















def id_from_url(link):
    if 'instagram.com/p/' not in link:
        # self.logger.error('Unexpected link')
        return False
    link = link.split('/')
    code = link[link.index('p') + 1]

    alphabet = {
        '-': 62, '1': 53, '0': 52, '3': 55, '2': 54, '5': 57, '4': 56,
        '7': 59, '6': 58, '9': 61, '8': 60, 'A': 0, 'C': 2, 'B': 1,
        'E': 4, 'D': 3, 'G': 6, 'F': 5, 'I': 8, 'H': 7, 'K': 10, 'J': 9,
        'M': 12, 'L': 11, 'O': 14, 'N': 13, 'Q': 16, 'P': 15, 'S': 18,
        'R': 17, 'U': 20, 'T': 19, 'W': 22, 'V': 21, 'Y': 24, 'X': 23,
        'Z': 25, '_': 63, 'a': 26, 'c': 28, 'b': 27, 'e': 30, 'd': 29,
        'g': 32, 'f': 31, 'i': 34, 'h': 33, 'k': 36, 'j': 35, 'm': 38,
        'l': 37, 'o': 40, 'n': 39, 'q': 42, 'p': 41, 's': 44, 'r': 43,
        'u': 46, 't': 45, 'w': 48, 'v': 47, 'y': 50, 'x': 49, 'z': 51,
        }

    result = 0
    for char in code:
        result = result * 64 + alphabet[char]
    return result


def url_from_id(id):
    id = int(id)
    alphabet = {
        '-': 62, '1': 53, '0': 52, '3': 55, '2': 54, '5': 57, '4': 56,
        '7': 59, '6': 58, '9': 61, '8': 60, 'A': 0, 'C': 2, 'B': 1,
        'E': 4, 'D': 3, 'G': 6, 'F': 5, 'I': 8, 'H': 7, 'K': 10, 'J': 9,
        'M': 12, 'L': 11, 'O': 14, 'N': 13, 'Q': 16, 'P': 15, 'S': 18,
        'R': 17, 'U': 20, 'T': 19, 'W': 22, 'V': 21, 'Y': 24, 'X': 23,
        'Z': 25, '_': 63, 'a': 26, 'c': 28, 'b': 27, 'e': 30, 'd': 29,
        'g': 32, 'f': 31, 'i': 34, 'h': 33, 'k': 36, 'j': 35, 'm': 38,
        'l': 37, 'o': 40, 'n': 39, 'q': 42, 'p': 41, 's': 44, 'r': 43,
        'u': 46, 't': 45, 'w': 48, 'v': 47, 'y': 50, 'x': 49, 'z': 51,
        }
    result = ''
    while id:
        id, char = id // 64, id % 64
        result += list(alphabet.keys())[list(alphabet.values()).index(char)]
    return 'https://instagram.com/p/' + result[::-1] + '/'



from typing import List
from funcy import rcompose, mapcat
import time
from ..bot import Bot
from ..nodes import User, Media
from .common import accepts, cycled_api_call


@accepts(User)
def following(bot, nodes,  args) -> List[User]:

    pack_user = lambda item: User(id=item['pk'], username=item['username'], data=item)
    amount = args.get('amount')

    process = rcompose(
        lambda user: user.id if user.id else user.get_id(bot),
        lambda id: cycled_api_call(amount, bot, bot.api.get_user_followings, id, 'users'),
        lambda gen: map(pack_user, gen)
    )

    result = mapcat(process, nodes)

    return result, bot.last
